from typing import Callable
import importlib
import functools
from wax.lessweb.utils import func_arg_spec
from pathlib import Path


class Importer:
    def __call__(self, package: str):
        return importlib.import_module(package)

    def __getattr__(self, item):
        return importlib.import_module(item)


lib = Importer()


def eval_func(lambda_func: str, **inject) -> Callable:
    kwargs = {}
    fn = eval(lambda_func)
    for arg_name in func_arg_spec(fn):
        if arg_name in inject:
            kwargs[arg_name] = inject[arg_name]
    return functools.partial(fn, **kwargs)


default_fn = {}
lib_path = Path('script')


def default_func(func_name: str, **inject) -> Callable:
    lambda_func = (lib_path / f'{func_name}.py').read_text(encoding='utf-8').strip()
    return eval_func(lambda_func, **inject)



from typing import Callable, Optional, Type, get_type_hints, Dict, Any, List, Tuple

from functools import lru_cache
from .context import Context, Request, Response
from .webapi import BadParamError
from .bridge import ParamStr
from .typehint import optional_core, generic_core, is_generic_type, get_origin
from .utils import func_arg_spec
from .storage import Storage


__all__ = ['request_bridge']


@lru_cache(maxsize=None)
def model_or_service(cls: Type) -> int:
    """
    :return: 1=Model 2=Service 0=None
    """
    try:
        for prop_type in get_type_hints(cls).values():
            if prop_type == int or prop_type == str:
                return 1
            elif prop_type == Context or prop_type == Request or prop_type == Response \
                    or model_or_service(prop_type) == 2:
                return 2
            else:
                return 1
        return 0
    except:
        return 0


def fetch_service(ctx: Context, service_type: Type):
    """
    :return:  cast(service_type, ctx)
    """
    params: Dict[str, Any] = {}
    for realname, realtype in Storage.type_hints(service_type).items():
        if realtype == Context:
            params[realname] = ctx
        elif realtype == Request:
            params[realname] = ctx.request
        elif realtype == Response:
            params[realname] = ctx.response
        elif model_or_service(realtype) == 2:
            params[realname] = fetch_service(ctx, realtype)
        else:
            pass  # 其他类型不注入
    service_obj = service_type()
    for key, val in params.items():
        setattr(service_obj, key, val)
    return service_obj


def request_bridge(inputval: Any, target_type: Type):
    """
    :return:  cast(target_type, inputval)
    """
    if target_type == Any:
        return inputval
    target_is_optional, target_type = optional_core(target_type)
    if inputval is None:
        if target_is_optional:
            return None
        else:
            raise ValueError("Cannot assign None when expected %s" % target_type)
    if isinstance(inputval, ParamStr):
        if issubclass(target_type, int):
            return target_type(int(inputval))
        else:
            return target_type(inputval)
    if isinstance(inputval, dict):
        if model_or_service(target_type) == 1:
            target_obj = target_type()
            for prop_name, prop_type in Storage.type_hints(target_type).items():
                if prop_name in inputval:
                    prop_value = request_bridge(inputval[prop_name], prop_type)
                    setattr(target_obj, prop_name, prop_value)
            return target_obj
        else:
            return target_type(**inputval)
    elif isinstance(inputval, list):
        if is_generic_type(target_type) and get_origin(target_type) == list:
            item_type = generic_core(target_type)
            return [request_bridge(item, item_type) for item in inputval]
        else:
            return target_type(*inputval)
    elif isinstance(inputval, target_type):
        return inputval
    else:
        return target_type(inputval)


def fetch_model(ctx: Context, target_type: Type) -> Any:
    if is_generic_type(target_type) and get_origin(target_type) == list:
        if not ctx.request.is_json():
            raise ValueError("Need JSON request when expected %s" % target_type)
        inputval = ctx.request.json_input
        if not isinstance(inputval, list):
            raise ValueError("Need JSON array request when expected %s" % target_type)
        item_type = generic_core(target_type)
        return [request_bridge(item, item_type) for item in inputval]
    else:
        target_obj = target_type()
        for realname, prop_type in Storage.type_hints(target_type).items():
            queryname = ctx.request._aliases.get(realname, realname)
            inputval = ctx.request.get_input(queryname)
            if inputval is not None:
                try:
                    setattr(target_obj, realname, request_bridge(inputval, prop_type))
                except Exception as e:
                    raise BadParamError(param=realname, message=str(e))
            else:
                pass  # 不赋值&不报错
        return target_obj


def fetch_param(ctx: Context, fn: Callable) -> Tuple[List, Dict[str, Any]]:
    """
    fn: dealer function
    return: Dict[realname, Context|Request|Response|Model|...]
    """
    args: List = []
    kwargs: Dict[str, Any] = {}
    for realname, (realtype, has_default, positional_only) in func_arg_spec(fn).items():
        value: Any
        if realtype == Context:
            value = ctx
        elif realtype == Request:
            value = ctx.request
        elif realtype == Response:
            value = ctx.response
        elif model_or_service(realtype) == 2:
            value = fetch_service(ctx, realtype)
        else:
            _, realtype = optional_core(realtype)
            if not isinstance(realtype, type) and not is_generic_type(realtype):
                raise BadParamError(param=realname, message='Unsupported type %s' % realtype)
            if positional_only:
                value = fetch_model(ctx, realtype)
            else:
                queryname = ctx.request._aliases.get(realname, realname)
                inputval = ctx.request.get_input(queryname)
                if inputval is not None:
                    try:
                        value = request_bridge(inputval, realtype)
                    except Exception as e:
                        raise BadParamError(param=realname, message=str(e))
                elif not has_default:
                    raise BadParamError(param=realname, message='Missing required param')
                else:
                    continue  # 不赋值&不报错
        # 将value追加到args或kwargs
        if positional_only:
            args.append(value)
        else:
            kwargs[realname] = value

    return args, kwargs



"""
Web application
(from lessweb)
"""
from datetime import datetime
import itertools
import json
import logging
import os
import re
import traceback
from types import GeneratorType
from typing import List, Any, Callable, Dict, Optional

from .webapi import BadParamError, NotFoundError, HttpStatus
from .webapi import http_methods
from .context import Context
from .model import fetch_param
from .storage import Storage
from .utils import eafp, re_standardize, makedir
from .bridge import make_response_encoder, JsonBridgeFunc
from .pluginproto import PluginProto


__all__ = [
    "Interceptor", "Mapping", "interceptor", "Application",
]


# Application.interceptors: List[Interceptor]
class Interceptor:
    """Interceptor to定义拦截器based on path prefix"""
    def __init__(self, pattern, method, dealer, patternobj) -> None:
        self.pattern: str = pattern
        self.method: str = method
        self.dealer: Callable = dealer
        self.patternobj: Any = patternobj


# Application.mapping: List[Mapping]
class Mapping:
    """Mapping to定义请求处理者和path的对应关系"""
    def __init__(self, pattern, method, dealer, doc, patternobj) -> None:
        self.pattern: str = pattern
        self.method: str = method
        self.dealer: Callable = dealer
        self.doc: str = doc
        self.patternobj: Any = patternobj


def build_controller(dealer):
    """
    把接收多个参数的dealer转变成只接收一个参数(ctx)的函数
    """
    def _1_controller(ctx:Context):
        try:
            args, params = fetch_param(ctx, dealer)
        except BadParamError:
            raise
        except Exception as e:
            raise BadParamError(message=str(e), param='')
        return dealer(*args, **params)

    return _1_controller


def interceptor(dealer):
    """
    为controller添加interceptor的decorator
    在dealer函数中调用ctx()，就会执行它修饰的controller
    """
    def _1_wrapper(fn):
        def _1_1_controller(ctx:Context):
            ctx.app_stack.append(build_controller(fn))
            args, params = fetch_param(ctx, dealer)
            result = dealer(*args, **params)
            ctx.app_stack.pop()  # 有多次调用ctx()的可能性，比如批量删除
            return result

        return _1_1_controller

    return _1_wrapper


class Application(object):
    """
    Application to delegate requests based on path.

    Example:

        from lessweb import Application
        app = Application()
        app.add_mapping('/hello', lambda ctx: 'Hello!')
        app.run(port=8080)

    """
    def __init__(self, encoding:str='utf-8') -> None:
        self.mapping: List[Mapping] = []
        self.interceptors: List[Interceptor] = []
        self.response_bridges: List[Callable] = []
        self.response_encoder: Any = make_response_encoder([])
        self.encoding: str = encoding
        self.plugins: List[PluginProto] = []

    def _handle_with_dealers(self, ctx: Context):
        def _1_mapping_match():
            supported_methods = []
            for mapping in self.mapping:
                _ = mapping.patternobj.search(ctx.request.path)
                if _:
                    if mapping.method == ctx.request.method or mapping.method == '*':
                        ctx.request.param_input.load_url(_.groupdict())
                        return mapping.dealer
                    elif mapping.method != 'OPTIONS':
                        supported_methods.append(mapping.method)
            # end: for
            raise NotFoundError(methods=supported_methods)

        try:
            f = build_controller(_1_mapping_match())
            if f is None: return ''
            for itr in self.interceptors:
                if itr.patternobj.search(ctx.request.path) and (itr.method == ctx.request.method or itr.method == '*'):
                    f = interceptor(itr.dealer)(f)
            return f(ctx)
        except BadParamError as e:
            ctx.response.set_status(HttpStatus.BadRequest)
            return {'message': e.message, 'param': e.param}
        except NotFoundError as e:
            if e.methods:
                ctx.response.send_allow_methods(e.methods)
                ctx.response.set_status(HttpStatus.MethodNotAllowed)
            else:
                ctx.response.set_status(HttpStatus.NotFound)
            return repr(e)

    def add_interceptor(self, pattern: str, method: str, dealer: Callable):
        """
        Example:

            from lessweb import Application
            app = Application()
            app.add_interceptor(lambda ctx: ctx() + ' world!')
            app.add_mapping('/hello', 'GET', lambda ctx: 'Hello')
            app.run()
        """
        assert isinstance(pattern, str), 'pattern:[{}] should be RegExp str'.format(pattern)
        method = method.upper()
        assert method == '*' or method in http_methods, 'Method:[{}] should be * or one of {}'.format(method, http_methods)
        patternobj = re.compile(re_standardize(pattern))
        self.interceptors.insert(0, Interceptor(pattern, method, dealer, patternobj))

    def add_json_bridge(self, bridge_func: JsonBridgeFunc):
        self.response_bridges.append(bridge_func)
        self.response_encoder = make_response_encoder(self.response_bridges)

    def add_mapping(self, pattern: str, method: str, dealer: Callable):
        """
        Example:

            from lessweb import Application
            def say_hello(ctx, name):
                return 'Hello %s!' % name
            app = Application()
            app.add_mapping('/hello/(?P<name>.+)', 'GET', say_hello)
            app.run()
        """
        assert isinstance(pattern, str), 'pattern:[{}] should be RegExp str'.format(pattern)
        method = method.upper()
        assert method == '*' or method in http_methods, 'Method:[{}] should be * or one of {}'.format(method, http_methods)
        patternobj = re.compile(re_standardize(pattern))
        self.mapping.append(Mapping(pattern, method, dealer, '', patternobj))

    # add_*_interceptor / add_*_mapping are generated by code below:
    """
    for m in ['CONNECT', 'DELETE', 'GET', 'HEAD', 'OPTIONS', 'PATCH', 'POST', 'PUT']:
        print(("def add_{m}_interceptor(self, pattern, dealer): return self.add_interceptor(pattern, '{M}', dealer)\n"
        "def add_{m}_mapping(self, pattern, dealer, view=None): return self.add_mapping(pattern, '{M}', dealer, view)\n")
        .format(m=m.lower(), M=m))
    """
    def add_connect_interceptor(self, pattern: str, dealer: Callable):
        return self.add_interceptor(pattern, 'CONNECT', dealer)

    def add_connect_mapping(self, pattern: str, dealer: Callable):
        return self.add_mapping(pattern, 'CONNECT', dealer)

    def add_delete_interceptor(self, pattern: str, dealer: Callable):
        return self.add_interceptor(pattern, 'DELETE', dealer)

    def add_delete_mapping(self, pattern: str, dealer: Callable):
        return self.add_mapping(pattern, 'DELETE', dealer)

    def add_get_interceptor(self, pattern: str, dealer: Callable):
        return self.add_interceptor(pattern, 'GET', dealer)

    def add_get_mapping(self, pattern: str, dealer: Callable):
        return self.add_mapping(pattern, 'GET', dealer)

    def add_head_interceptor(self, pattern: str, dealer: Callable):
        return self.add_interceptor(pattern, 'HEAD', dealer)

    def add_head_mapping(self, pattern: str, dealer: Callable):
        return self.add_mapping(pattern, 'HEAD', dealer)

    def add_options_interceptor(self, pattern: str, dealer: Callable):
        return self.add_interceptor(pattern, 'OPTIONS', dealer)

    def add_options_mapping(self, pattern: str, dealer: Callable):
        return self.add_mapping(pattern, 'OPTIONS', dealer)

    def add_patch_interceptor(self, pattern: str, dealer: Callable):
        return self.add_interceptor(pattern, 'PATCH', dealer)

    def add_patch_mapping(self, pattern: str, dealer: Callable):
        return self.add_mapping(pattern, 'PATCH', dealer)

    def add_post_interceptor(self, pattern: str, dealer: Callable):
        return self.add_interceptor(pattern, 'POST', dealer)

    def add_post_mapping(self, pattern: str, dealer: Callable):
        return self.add_mapping(pattern, 'POST', dealer)

    def add_put_interceptor(self, pattern: str, dealer: Callable):
        return self.add_interceptor(pattern, 'PUT', dealer)

    def add_put_mapping(self, pattern: str, dealer: Callable):
        return self.add_mapping(pattern, 'PUT', dealer)

    def add_plugin(self, plugin: PluginProto):
        self.plugins.append(plugin)
        plugin.init_app(self)

    def wsgifunc(self, *middleware):
        """
            Example:

                import lessweb
                app = lessweb.Application()
                app.add_interceptor('/', '*', lambda ctx: ctx() + ' world!')
                app.add_mapping('/hello', lambda ctx: 'Hello')
                application = app.wsgifunc()
        """
        def wsgi(env, start_resp):
            def _1_peep(iterator):
                """Peeps into an iterator by doing an iteration
                and returns an equivalent iterator.
                """
                # wsgi requires the headers first
                # so we need to do an iteration
                # and save the result for later
                try:
                    firstchunk = next(iterator)
                except StopIteration:
                    firstchunk = ''
                return itertools.chain([firstchunk], iterator)

            ctx = Context(self)
            ctx.request.load(env)
            try:
                mimekey = 'html'
                resp = self._handle_with_dealers(ctx)
                resp_content_type = ctx.response.get_header('Content-Type')
                if isinstance(resp, GeneratorType):
                    result = _1_peep(resp)
                else:
                    if not isinstance(resp, (bytes, str)) and resp is not None:
                        if not resp_content_type or \
                                (resp_content_type and 'json' in resp_content_type.lower()):
                            resp = json.dumps(resp, ensure_ascii=False, cls=self.response_encoder)
                            mimekey = 'json'
                        else:
                            resp = str(resp)
                    result = (resp,)
                if not resp_content_type:
                    ctx.response.send_content_type(mimekey=mimekey, encoding=self.encoding)
            except Exception as e:
                logging.exception(e)
                ctx.response.send_content_type(encoding=self.encoding)
                ctx.response.set_status(HttpStatus.InternalServerError)
                result = (traceback.format_exc(),)

            def _2_build_result(result):
                for r in result:
                    if isinstance(r, bytes):
                        yield r
                    elif isinstance(r, str):
                        yield r.encode(self.encoding, 'replace')
                    elif r is None:
                        yield b''
                    else:
                        yield str(r).encode(self.encoding)

            result = _2_build_result(result)
            status_wrap = ctx.response.get_status()
            status_core = status_wrap.value if isinstance(status_wrap, HttpStatus) else status_wrap
            status_text = f'{status_core.code} {status_core.reason}'
            headers = list(ctx.response._headers.items())
            for cookie in ctx.response._cookies.values():
                headers.append(('Set-Cookie', cookie.dumps()))
            start_resp(status_text, headers)
            return itertools.chain(result, (b'',))

        for m in middleware:
            wsgi = m(wsgi)

        return wsgi

    def run(self, wsgifunc=None, port:int=8080, homepath:str='', staticpath:Optional[str]='static'):
        """
        Example:

            from lessweb import Application
            app = Application()
            app.add_interceptor('/', '*', lambda ctx: ctx() + ' world!')
            app.add_mapping('/hello', lambda ctx: 'Hello')
            app.run(port=80, homepath='/api')
        """
        from aiohttp import web
        from aiohttp_wsgi import WSGIHandler  # type: ignore
        app = web.Application()
        if wsgifunc is None:
            wsgifunc = self.wsgifunc()

        if homepath.endswith('/'):
            homepath = homepath[:-1]
        if homepath and homepath[0] != '/':
            homepath = '/' + homepath

        if staticpath is not None:
            makedir(staticpath)
            app.router.add_static(prefix='/static/', path=staticpath)
        app.router.add_route("*", homepath + "/{path_info:.*}", WSGIHandler(wsgifunc))
        web.run_app(app, port=port)



from typing import Tuple, Dict, List
import json
import jsonschema
import base64
import re
from wax.lessweb import Request, Response, BadParamError
from wax.lessweb.utils import re_standardize, eafp
from wax.lessweb.webapi import http_methods, NotFoundError, HttpStatus
from wax.load_config import config
from wax.load_swagger import SwaggerData
from wax.load_func import eval_func, default_func
from wax.service import StateServ


def base64ed(buf: bytes) -> str:
    return base64.standard_b64encode(buf).decode()


def opt_number(n):
    try:
        return int(n)
    except:
        try:
            return float(n)
        except:
            return n


def validate(instance, *, schema) -> str:
    try:
        jsonschema.validate(instance=instance, schema=schema, resolver=SwaggerData.resolver, format_checker=jsonschema.draft7_format_checker)
        return ''
    except jsonschema.ValidationError as e:
        return str(e)


class InternalError(Exception):
    pass


def hit_endpoint(realpath: str) -> Tuple[Dict, Dict]:
    """
    :return (endpoint, url_params)
    """
    path = realpath[len(config['mockapi-prefix']):]
    swagger_data = SwaggerData.get()
    for endpoint_path, endpoint in swagger_data['paths'].items():
        search_ret = re.compile(re_standardize(endpoint_path)).search(path)
        if search_ret:
            return endpoint, search_ret.groupdict()
    raise NotFoundError(methods=[])


def hit_operation(endpoint: Dict, method: str) -> Dict:
    supported_methods = []
    for op_method, operation in endpoint.items():
        op_method = op_method.upper()
        if op_method not in http_methods:
            continue
        supported_methods.append(op_method)
        if op_method == method.upper():
            return operation
    raise NotFoundError(methods=supported_methods)


def check_param_str(param_name, *, param_value, schema) -> None:
    """
    检查通过则return，不通过则返回BadParamError
    """
    error_message = validate(param_value, schema=schema)
    if error_message:
        param_value = opt_number(param_value)
        if validate(param_value, schema=schema):
            raise BadParamError(param=param_name, message=error_message)


def param_check(params: List[Dict], request: Request, url_params: Dict) -> None:
    """
    检查通过则return，不通过则返回BadParamError
    """
    for param_dict in params:
        param_name = param_dict['name']
        param_in = param_dict.get('in', 'query')
        param_required = param_dict.get('required', False)
        if param_in == 'header':
            if not request.contains_header(param_name):
                if param_required:
                    raise BadParamError(param=param_name, message='header param is required')
                else:
                    continue
            param_value = request.get_header(param_name)
        elif param_in == 'cookie':
            if not request.contains_cookie(param_name):
                if param_required:
                    raise BadParamError(param=param_name, message='cookie param is required')
                else:
                    continue
            param_value = request.get_cookie(param_name)
        elif param_in == 'path':
            if param_name not in url_params:
                if param_required:
                    raise BadParamError(param=param_name, message='path param is required')
                else:
                    continue
            param_value = url_params[param_name]
        else:
            query_params = request.param_input.query_input
            if param_name not in query_params:
                if param_required:
                    raise BadParamError(param=param_name, message='query param is required')
                else:
                    continue
            param_value = query_params[param_name][0]
        if not param_dict.get('allowEmptyValue', True) and not param_value:
            raise BadParamError(param=param_name, message='allowEmptyValue is true but param is empty')
        check_param_str(param_name, param_value=param_value, schema=param_dict['schema'])


def body_check(operation: Dict, request: Request) -> None:
    """
    检查通过则return，不通过则返回BadParamError
    """
    content_dict = eafp(lambda: operation['requestBody']['content'], {})
    real_content_type = request.get_content_type().lower()
    for content_key, content_val in content_dict.items():
        if content_key.lower() in real_content_type:
            schema = content_val['schema']
            if request.is_json():
                error_message = validate(request.json_input, schema=schema)
            elif request.is_form():
                form_dict = {key: val[0] for key, val in request.param_input.form_input.items()}
                for key, val in request.file_input.items():
                    form_dict = base64ed(val[0].value)
                error_message = validate(form_dict, schema=schema)
                if not error_message:
                    return
                error_message = validate(
                    {key: opt_number(val) for key, val in form_dict.items()}, schema=schema)
            else:
                error_message = validate(base64ed(request.body_data or b''), schema=schema)
            if error_message:
                raise BadParamError(message=error_message, param='requestBody')
            else:
                return
    raise BadParamError(message='unsupported request content-type', param=real_content_type)


def hit_example(operation: Dict, state) -> Tuple[Dict, str]:
    """
    :return (schema, example_json)
    """
    state_basic = state.basic.get()
    if state_basic:
        state_status_code, state_content_type, state_example_name = state_basic.split(':', 2)
    else:
        state_status_code = state_content_type = state_example_name = ''
    if not operation.get('responses'):
        raise InternalError('responses is empty')
    for status_code, response_val in operation['responses'].items():
        if not state_status_code or state_status_code == status_code:
            for content_key, content_val in response_val['content'].items():
                if not state_content_type or state_content_type == content_key:
                    for example_name, example_val in content_val.get('examples', {}).items():
                        if not state_example_name or state_example_name == example_name:
                            return content_val['schema'], json.dumps(example_val['value'])
                    else:
                        raise InternalError('response fail to match example')
            else:
                raise InternalError('response fail to match content-type')
    else:
        raise InternalError('response fail to match status_code')


def rotate_fetch(arr: List):
    if not arr:
        return None
    ret = arr.pop(0)
    arr.append(ret)
    return ret


def json_egg(obj, *, array):
    """
    array is False:
      basic -> basic
      list -> list[0], rotate(list, 1)
      dict -> {key: json_egg(val)}
    array is True:
      list -> list
      some -> [some]
    array is None:
      None
    array == 0:
      []
    array > 0:
      list -> list[:n], rotate(list, n)
      some -> [some] * n
    array < 0:
      list -> list[n-1:-1:-1]
      some -> [some] * (-n)
    array is list:
      n := list[0], rotate(array, 1)
      assert n is (int | None)
      list -> json_egg(list, array=n)
    """
    if array is False:
        if isinstance(obj, list):
            return rotate_fetch(obj)
        elif isinstance(obj, dict):
            ret = {}
            for key, val in obj.items():
                if '[]' in key: continue
                dual_key = key + '[]'
                if dual_key in obj:
                    ret[key] = json_egg(val, array=obj[dual_key])
                else:
                    ret[key] = json_egg(val, array=False)
            return ret
        else:
            return obj
    elif array is True:
        if isinstance(obj, list):
            return obj
        else:
            return [obj]
    elif isinstance(array, int) and array is None:
        return None
    elif isinstance(array, int) and array == 0:
        return []
    elif isinstance(array, int) and array > 0:
        if isinstance(obj, list):
            return [rotate_fetch(obj) for _ in range(array)]
        else:
            return [json_egg(obj, array=False) for _ in range(array)]
    elif isinstance(array, int) and array < 0:
        if isinstance(obj, list):
            ret = [rotate_fetch(obj) for _ in range(-array)]
        else:
            ret = [json_egg(obj, array=False) for _ in range(-array)]
        return ret.reverse() or ret
    elif isinstance(array, list):
        if not array:
            return None
        n = rotate_fetch(array)
        if not (n is None or (isinstance(n, int) and n >= 0)):
            raise InternalError(f'Invalid example: {n} in {array}')
        return json_egg(obj, array=n)
    else:
        raise InternalError(f'Invalid example: {array}')


def chain_filter(func_chain: List[str], resp_obj, request, response, state):
    for func_item in func_chain:
        if func_item.startswith('@'):
            func = default_func(func_item[1:], request=request, response=response, state=state)
        else:
            func = eval_func(func_item, request=request, response=response, state=state)
        resp_obj = func(resp_obj)
    return resp_obj


def response_check(schema, resp_obj) -> None:
    """
    检查通过则return，不通过则返回InternalError
    """
    error_message = validate(resp_obj, schema=schema)
    if error_message:
        error_message += '\n\n-----------------\nresponse:\n'
        try:
            error_message += json.dumps(resp_obj, ensure_ascii=False)
        except:
            error_message += str(schema)
        raise InternalError(error_message)


def mock_dealer(request: Request, response: Response, state: StateServ):
    try:
        endpoint, url_params = hit_endpoint(request.path)
        if 'parameters' in endpoint:
            param_check(endpoint['parameters'], request=request, url_params=url_params)
        operation = hit_operation(endpoint, method=request.method)
        params = operation.get('parameters', [])
        param_check(params, request=request, url_params=url_params)
        if operation.get('requestBody'):
            body_check(operation, request=request)
        state.operation_id = operation['operationId']
        schema, example_json = hit_example(operation, state=state)
        example = json.loads(example_json)
        func_chain = example.get('@', [])
        resp_obj = json_egg(example, array=False)
        if isinstance(resp_obj, dict) and '$' in resp_obj:
            resp_obj = resp_obj['$']
        resp_obj = chain_filter(func_chain, resp_obj=resp_obj,
                                request=request, response=response, state=state)
        response_check(schema=schema, resp_obj=resp_obj)
        return resp_obj
    except InternalError as e:
        response.send_content_type(mimekey='txt', encoding='utf-8')
        response.set_status(HttpStatus.InternalServerError)
        return str(e)


from typing import Dict, Any
import json
import jsonschema
from os.path import getmtime


class SwaggerData:
    json_path = ''
    redis_prefix = 'waxapi::'
    swagger_data: Dict = {}
    resolver = None
    last_modify = 0

    @classmethod
    def init(cls, json_path):
        cls.json_path = json_path
        cls.last_modify = getmtime(json_path)
        cls.swagger_data = json.loads(open(json_path).read())
        cls.resolver = jsonschema.RefResolver.from_schema(cls.swagger_data)
        cls.redis_prefix = 'waxapi::' + cls.swagger_data['info']['title'] + '::' + cls.swagger_data['info']['version'] + '::'

    @classmethod
    def get(cls) -> Dict:
        last_modify = getmtime(cls.json_path)
        if last_modify != cls.last_modify:
            cls.swagger_data = json.loads(open(cls.json_path).read())
            cls.resolver = jsonschema.RefResolver.from_schema(cls.swagger_data)
            cls.last_modify = last_modify
        return cls.swagger_data



"""lessweb: 用最python3的方法创建web apps"""


__version__ = '0.3.1'
__author__ = [
    'qorzj <inull@qq.com>',
]

__license__ = "MIT"

# from . import application, context, model, storage, webapi

from .application import interceptor, Application
from .context import Context, Request, Response
from .storage import Storage
from .bridge import uint, ParamStr, MultipartFile, Jsonizable
from .webapi import BadParamError, NotFoundError, Cookie, HttpStatus, ResponseStatus
from .utils import _nil, eafp



from typing import Dict, List
from mako.template import Template  # type: ignore
from wax.lessweb import BadParamError
from wax.lessweb.webapi import http_methods
from wax.service import StateServ
from wax.load_config import config
from wax.load_swagger import SwaggerData


def split_tag(tag: str) -> List[str]:
    assert tag
    segs = tag.replace('/', '-').split('-', 2)
    if len(segs) == 1:
        segs.append('API')
    if len(segs) == 2:
        segs.append('API')
    return segs


class Operation:
    method: str
    path: str
    summary: str
    operationId: str
    description: str
    # state
    basic: str
    extra: str
    all_example: List[str]  # List[stateBasic]


# {majorTag: {dirTag: {menuTag: List[Operation]}}}
tag_tree: Dict[str, Dict[str, Dict[str, List[Operation]]]] = {}
op_index: Dict[str, Operation] = {}  # {operationId: Operation}


def load_tag():
    swagger_data = SwaggerData.get()
    for endpoint_path, endpoint in swagger_data['paths'].items():
        for op_method, operation in endpoint.items():
            if op_method.upper() not in http_methods: continue
            for tag in operation.get('tags', []):
                if not tag:
                    continue
                major_tag, dir_tag, menu_tag = split_tag(tag)
                tag_tree.setdefault(major_tag, {})
                tag_tree[major_tag].setdefault(dir_tag, {})
                tag_tree[major_tag][dir_tag].setdefault(menu_tag, [])
                op = Operation()
                op.method = op_method.upper()
                op.path = endpoint_path
                op.summary = operation['summary']
                op.operationId = operation['operationId']
                op.description = operation.get('description', '')
                op.all_example = []
                for status_code, response_val in operation['responses'].items():
                    for content_key, content_val in response_val['content'].items():
                        for example_name, _ in content_val.get('examples', {}).items():
                            op.all_example.append(f'{status_code}:{content_key}:{example_name}')
                tag_tree[major_tag][dir_tag][menu_tag].append(op)
                op_index[op.operationId] = op


load_tag()


def operation_list(tag: str=''):
    if not tag and tag_tree:
        tag = list(tag_tree.keys())[0]
    major_tag, dir_tag, menu_tag = split_tag(tag)
    return Template(filename='wax-www/tpl/tag_list_page.mako', input_encoding='utf-8', output_encoding='utf-8').render(
        tag_tree=tag_tree,
        major_tag=major_tag,
        dir_tag=dir_tag,
        menu_tag=menu_tag,
        git_url=config['git-url']
    )


def operation_detail(opId: str, state: StateServ):
    op = op_index.get(opId)
    if not op:
        raise BadParamError(message='opId不存在', param='opId')
    state.operation_id = opId
    op.basic = state.basic.get()
    op.extra = state.extra.get()
    return op


def operation_edit_state(state: StateServ, opId: str, basic:str='', extra:str=''):
    if opId not in op_index:
        raise BadParamError(message='opId不存在', param='opId')
    state.operation_id = opId
    if basic is not None:
        state.basic.set(basic, ex=86400 * 90)
    if extra is not None:
        state.extra.set(extra, ex=86400 * 90)
    return {}



# coding: utf-8
"""
wax-mock
~~~~~~~~
Mock server with Swagger OpenAPI3 JSON
Setup
-----
.. code-block:: bash
    > pip install wax-mock
    > wax swagger_api.json

"""

from setuptools import setup, find_packages  # Always prefer setuptools over distutils
from codecs import open  # To use a consistent encoding
from os import path
from setuptools.command.install import install
import re
import ast

_version_re = re.compile(r'__version__\s+=\s+(.*)')
version = str(ast.literal_eval(
    _version_re.search(
        open('wax/__init__.py').read()
    ).group(1)
))
here = path.abspath(path.dirname(__file__))


class MyInstall(install):
    def run(self):
        print("-- installing... --")
        install.run(self)

setup(
        name = 'wax-mock',
        version=version,
        description='Mock server with Swagger OpenAPI3 JSON',
        long_description='\npip install wax-mock\n\n'
                         'wax swagger_api.json',
        url='https://pypi.python.org/project/wax-mock',
        author='qorzj',
        author_email='inull@qq.com',
        license='MIT',
        platforms=['any'],

        classifiers=[
            ],
        keywords='mock swagger openapi openapi3',
        packages = ['wax', 'wax.lessweb', 'wax.lessweb.plugin'],
        install_requires=[
            'jsonschema',
            'strict-rfc3339',
            'rfc3987',
            'aiohttp',
            'aiohttp_wsgi',
            'requests',
            'typing_inspect',
            'typing_extensions',
            'redis',
            'hiredis',
            'Mako',
        ],
        cmdclass={'install': MyInstall},
        entry_points={
            'console_scripts': [
                'wax = wax.main:entrypoint'
            ],
        },
    )
